# from typstwriter import compiler
# from qtpy import QtCore
# import time
# import pytest


# def mock_process_start


# class TestCompilerConnector_FS_onDemand:
#
#     # @pytest.fixture(scope="class")
#     # def patch(self, monkeypatch):
#     #     """Patch the QtCore.QProcess.start() function to do nothing."""
#     #     monkeypatch.setattr(QtCore.QProcess, "start", lambda: None)
#
#     def test_successful_compilation(self, tmp_path, qtbot):
#         """Test a successful compilation"""
#         fin = str(tmp_path / "fin")
#         fout = str(tmp_path / "fout")
#
#         c = compiler.CompilerConnector_FS_onDemand(fin=fin, fout=fout)
#
#         with qtbot.waitSignal(c.started):
#             c.start()

#
# class TestCompilerConnector_FS_live:
#     pass
#
#
# class TestWrappedCompilerConnector:
#     pass

